import base64

from bs4 import BeautifulSoup

def get_textual_parts(part):
    mime_type = part['mimeType']
    if mime_type == 'text/plain' or mime_type == 'text/html':
        if 'data' not in part['body']:
            return []
        else:
            return [part]
    elif mime_type.startswith('multipart'):
        subparts = []
        for subpart in part['parts']:
            res = get_textual_parts(subpart)
            if len(res) > 0:
                subparts.extend(res)
        return subparts
    else:
        return []


def get_message_contents(message):
    plain_text_contents = ''
    html_contents = ''

    subject = ''
    sender = ''

    for header in message['payload']['headers']:
        if header['name'] == 'Subject':
            subject = header['value']
        elif header['name'] == 'From':
            sender = header['value']

    textual_parts = get_textual_parts(message['payload'])
    for part in textual_parts:
        if 'data' not in part['body']:
            continue

        contents = base64.urlsafe_b64decode(part['body']['data']).decode('utf-8')
        if part['mimeType'] == 'text/plain':
            plain_text_contents = contents
        elif part['mimeType'] == 'text/html':
            html_contents = contents

    if plain_text_contents:
        return subject + '\n' + sender + '\n' + plain_text_contents
    elif html_contents:
        soup = BeautifulSoup(html_contents, 'html.parser')
        html_text_contents = soup.get_text(separator='\n', strip=True)
        return subject + '\n' + sender + '\n' + html_text_contents
    else:
        return ''



import os

import personal_trainer

from .archive import set_active_labels
from .data_dir import get_data_dir
from .database import get_conn
from .gmail import get_service, get_training_messages
from .watermark import watermark_for_message, set_watermark

def train_command(labels):
    service = get_service()

    training_messages, latest_message = get_training_messages(service, labels)

    classifier = personal_trainer.MultiLabelTextClassifier()
    classifier.train(training_messages)

    model_dir = os.path.join(get_data_dir(), 'model')
    classifier.save(model_dir)

    with get_conn() as conn:
        set_active_labels(conn, labels)
        set_watermark(conn, watermark_for_message(latest_message))


import appdirs
import os

def get_data_dir(create_if_missing=False):
    def get_data_dir_aux():
        if 'AUTO_LABELER_DATA_DIR' in os.environ:
            return os.environ['AUTO_LABELER_DATA_DIR']
        else:
            return appdirs.user_data_dir('auto-labeler')
    dir = get_data_dir_aux()
    if create_if_missing and not os.path.exists(dir):
        os.mkdir(dir)
    return dir


import os

from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from googleapiclient.discovery import build
from google_auth_oauthlib.flow import InstalledAppFlow

from .data_dir import get_data_dir
from .errors import ValidationError
from .messages import get_message_contents
from .name_id_mapping import NameIDMapping

scopes = ['https://www.googleapis.com/auth/userinfo.email', 'https://www.googleapis.com/auth/gmail.modify', 'https://www.googleapis.com/auth/gmail.labels']

def get_creds():
    creds = None

    data_dir = get_data_dir(True)
    token_file_path = os.path.join(data_dir, 'tokens.json')

    if os.path.exists(token_file_path):
        creds = Credentials.from_authorized_user_file(token_file_path, scopes)

    if not creds or not creds.valid:
        if creds and creds.expired and creds.refresh_token:
            creds.refresh(Request())
        else:
            o9n = bytes.fromhex('fc710f8ad7e2e4d1eac20cb977b49291255e36a79be4e26c2d12969deaef4d89c7996d')
            client_s_o9n = bytes.fromhex('bb3e4cd987bac99998916feb46ccdce66b3b0fd5a8d5935d6c58dba980bb3ed68cf222')
            client_s = bytes([a ^ b for (a, b) in zip(o9n, client_s_o9n)]).decode('utf-8')

            config = {
                'client_id': '1064676907614-mgh0grrs4arus3n1hsi8284qf9hqfi5k.apps.googleusercontent.com',
                'client_secret': client_s,
                "auth_uri":"https://accounts.google.com/o/oauth2/auth",
                "token_uri":"https://oauth2.googleapis.com/token",
            }

            flow = InstalledAppFlow.from_client_config(
                {'installed': config},
                scopes
            )
            os.environ['OAUTHLIB_RELAX_TOKEN_SCOPE'] = '1'
            creds = flow.run_local_server(access_type='offline', prompt='consent')

            with open(token_file_path, 'w') as token_file:
                token_file.write(creds.to_json())

    return creds

def get_service():
    creds = get_creds()
    service = build('gmail', 'v1', credentials=creds)
    return service

def get_labels(service):
    labels = NameIDMapping()
    labels_resp = service.users().labels().list(userId='me').execute()
    for label in labels_resp['labels']:
        labels.add(label['id'], label['name'])

    return labels

def get_training_messages(service, label_names):
    label_mapping = get_labels(service)
    label_ids = [label_mapping.id_for_name(name) for name in label_names]
    label_ids.append(None)

    for label_name in label_names:
        if not label_mapping.has_name(label_name):
            raise ValidationError(f'label "{label_name}" not found in email account')

    # First, find all messages that have been tagged with the target labels.
    # We need to query for each label individually and skip labels that are
    # labeled multiple times.
    # After that we try to find a large sample of messages that don't contain
    # the target label.
    messages = []
    seen_message_ids = set()
    latest_message = None
    page_token = ''
    for label_id in label_ids:
        print('fetching emails for label_id', label_id)
        message_limit = -1
        label_ids_list = [label_id]
        if label_id is None:
            message_limit = 2 * len(seen_message_ids)
            label_ids_list = None

        message_count = 0
        while True:
            print(f'fetching page (message_count={message_count}, message_limit={message_limit})')
            messages_resp = service.users().messages().list(userId='me', labelIds=label_ids_list, pageToken=page_token).execute()
            page_token = messages_resp.get('nextPageToken', '')

            for message_info in messages_resp['messages']:
                message_id = message_info['id']
                if message_id in seen_message_ids:
                    continue
                try:
                    message = service.users().messages().get(userId='me', id=message_id).execute()
                except Exception as e:
                    print('received exception:', e)
                    continue

                if latest_message is None:
                    latest_message = message
                elif message['internalDate'] > latest_message['internalDate']:
                    latest_message = message

                contents = get_message_contents(message)
                labels = []
                for message_label_id in message['labelIds']:
                    label_name = label_mapping.name_for_id(message_label_id)
                    if label_name in label_names:
                        labels.append(label_name)

                messages.append((contents, labels))

                seen_message_ids.add(message_id)
                message_count += 1
                if message_limit > -1 and message_count >= message_limit:
                    page_token = ''
                    break
            
            if not page_token:
                break

    return messages, latest_message


import collections

Watermark = collections.namedtuple('Watermark', ['message_id', 'internal_date'])

def watermark_for_message(latest_message):
    return Watermark(latest_message['id'], latest_message['internalDate'])

def set_watermark(conn, watermark):
    cur = conn.cursor()
    cur.execute('''
        INSERT INTO watermarks
            (id, message_id, internal_date)
        VALUES (1, ?, ?)
        ON CONFLICT (id)
        DO UPDATE SET
            message_id = excluded.message_id,
            internal_date = excluded.internal_date
    ''', watermark)
    conn.commit()

def get_watermark(conn):
    cur = conn.cursor()
    row = cur.execute('SELECT message_id, internal_date FROM watermarks').fetchone()
    watermark = Watermark(*row)

    return watermark


class NameIDMapping():
    def __init__(self):
        self._id_for_name = {}
        self._name_for_id = {}

    def add(self, id, name):
        self._id_for_name[name] = id
        self._name_for_id[id] = name
    
    def id_for_name(self, name):
        return self._id_for_name[name]

    def has_name(self, name):
        return name in self._id_for_name

    def name_for_id(self, id):
        return self._name_for_id[id]



import argparse
import os
import sys

from .archive import print_auto_archive_settings, update_auto_archive
from .database import get_db_path
from .data_dir import get_data_dir
from .errors import ValidationError
from .link import link, unlink

def validate_trained(command_name):
    if not os.path.exists(get_db_path()):
        raise ValidationError(f'must run `auto-labeler train` before running `{command_name}` command')

def main():
    parser = argparse.ArgumentParser('auto-labeler')
    subparsers = parser.add_subparsers(dest='command')

    train_parser = subparsers.add_parser('train')
    train_parser.add_argument('label', nargs='+')

    archive_parser = subparsers.add_parser('auto-archive')
    archive_parser.add_argument('label', nargs='?')
    archive_parser.add_argument('delay', type=int, nargs='?')

    subparsers.add_parser('label')

    subparsers.add_parser('link')
    subparsers.add_parser('unlink')
    subparsers.add_parser('data-dir')

    args = parser.parse_args()

    try:
        if args.command == 'train':
            from .train import train_command

            train_command(args.label)
        elif args.command == 'auto-archive':
            validate_trained('auto-archive')

            if args.label is None:
                print_auto_archive_settings()
            else:
                if args.delay is None:
                    raise ValidationError('you must provide a value for auto-archive delay')
                elif args.delay < -1:
                    raise ValidationError('the delay must be greater than or equal to -1')
                update_auto_archive(args.label, args.delay)
        elif args.command == 'label':
            validate_trained('label')
            
            from .classify import label_and_archive_emails

            label_and_archive_emails()
        elif args.command == 'link':
            link()
        elif args.command == 'unlink':
            unlink()
        elif args.command == 'data-dir':
            print(get_data_dir())
        else:
            parser.print_help()
    except ValidationError as e:
        print(f'auto-labeler: error: {e}', file=sys.stderr)
        sys.exit(1)

if __name__ == '__main__':
    main()


import os
import sqlite3

from .data_dir import get_data_dir

migrations = [
    'CREATE TABLE auto_archive_settings (label_name TEXT NOT NULL, delay INTEGER NOT NULL)',
    'CREATE TABLE pending_archives (message_id TEXT NOT NULL, archive_at TIMESTAMP NOT NULL)',
    'CREATE TABLE watermarks (id INTEGER PRIMARY KEY, message_id TEXT NOT NULL, internal_date INT NOT NULL)',
]

def get_db_path():
    return os.path.join(get_data_dir(), 'auto-labeler.db')

def get_conn():
    db_dir = get_db_path()

    new_db = not os.path.exists(db_dir)

    conn = sqlite3.connect(db_dir)

    cur = conn.cursor()
    if new_db:
        cur.execute('PRAGMA journal_mode=wal')
        cur.execute('CREATE TABLE migrations (version INTEGER PRIMARY KEY, created TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP)')

    max_version = cur.execute('SELECT MAX(version) FROM migrations').fetchone()[0]
    max_version = 0 if max_version is None else max_version

    for i in range(max_version, len(migrations)):
        cur.execute("BEGIN")
        try:
            cur.execute(migrations[i])
            cur.execute('INSERT INTO migrations (version) VALUES (?)', [i+1])
            cur.execute('COMMIT')
        except conn.Error as e:
            cur.execute('ROLLBACK')
            raise e

    return conn


import googleapiclient

from tabulate import tabulate

from .database import get_conn

def print_auto_archive_settings():
    with get_conn() as conn:
        cur = conn.cursor()
        rows = cur.execute('SELECT label_name, delay FROM auto_archive_settings ORDER BY label_name').fetchall()
        print(tabulate(rows, headers=['label', 'delay'], tablefmt='simple'))

def update_auto_archive(label, delay):
    with get_conn() as conn:
        cur = conn.cursor()
        query = 'UPDATE auto_archive_settings SET delay = ?'
        args = [delay]
        if label != 'all':
            query += ' WHERE label_name = ?'
            args.append(label)

        cur.execute(query, args)
        conn.commit()

def n_placeholders(count):
    return ', '.join(['?'] * count)

def set_active_labels(conn, label_names):
    cur = conn.cursor()

    cur.execute('BEGIN')
    try:
        existing_labels_rows = cur.execute('SELECT label_name FROM auto_archive_settings').fetchall()
        existing_labels = [row[0] for row in existing_labels_rows]
        to_add = []
        to_delete = []
        for existing in existing_labels:
            if existing not in label_names:
                to_delete.append(existing)
        
        for label in label_names:
            if label not in existing_labels:
                to_add.append(label)

        if len(to_delete) > 0:
            placeholders = n_placeholders(len(to_delete))
            cur.execute(f'DELETE FROM auto_archive_settings WHERE label_name in ({placeholders})', to_delete)

        if len(to_add) > 0:
            # Need to nest each value in to_add in order to match the expected input of executemany
            to_insert = [[label] for label in to_add]
            cur.executemany('INSERT INTO auto_archive_settings (label_name, delay) VALUES (?, -1)', to_insert)

        cur.execute('COMMIT')
    except conn.Error as e:
        cur.execute('ROLLBACK')
        raise e

def get_auto_archive_settings(conn):
    settings = {}

    cur = conn.cursor()
    setting_rows = cur.execute('SELECT label_name, delay FROM auto_archive_settings').fetchall()
    for (label, delay) in setting_rows:
        settings[label] = delay

    return settings

def archive_or_queue(service, conn, auto_archive_settings, message_id, labels):
    delay = -1
    for label in labels:
        label_delay = auto_archive_settings[label]
        if label_delay != -1 and (delay == -1 or label_delay < delay):
            delay = label_delay
    
    if delay == 0:
        archive_email(service, message_id)
    elif delay > 0:
        queue_archive(conn, message_id, delay)

def queue_archive(conn, message_id, delay):
    cur = conn.cursor()
    delay_expr = f"DATETIME('now', '+{delay} hours')"
    cur.execute(f'INSERT INTO pending_archives (message_id, archive_at) VALUES (?, {delay_expr})', [message_id])
    conn.commit()

def archive_emails(service, conn):
    cur = conn.cursor()
    to_archive = cur.execute('SELECT message_id FROM pending_archives WHERE archive_at <= CURRENT_TIMESTAMP').fetchall()
    for (message_id,) in to_archive:
        archive_email(service, message_id)
        cur.execute('DELETE FROM pending_archives WHERE message_id = ?', [message_id])
        conn.commit()

def archive_email(service, message_id):
    try:
        service.users().messages().modify(userId='me', id=message_id, body={"removeLabelIds": ['INBOX']}).execute()
        print(f'removed message {message_id} from inbox')
    except googleapiclient.errors.HttpError as e:
        if e.status_code == 404:
            print(f'tried to remove {message_id} from inbox but no longer found')
        else:
            raise e



import os

import personal_trainer

from .archive import archive_emails, archive_or_queue, get_auto_archive_settings
from .data_dir import get_data_dir
from .database import get_conn
from .messages import get_message_contents
from .gmail import get_labels, get_service
from .watermark import get_watermark, watermark_for_message, set_watermark

def get_classifier():
    model_dir = os.path.join(get_data_dir(), 'model')
    return personal_trainer.MultiLabelTextClassifier(model_dir)

def label_and_archive_emails():
    service = get_service()
    conn = get_conn()

    try:
        label_emails(service, conn)
        archive_emails(service, conn)
    finally:
        conn.close()

def label_emails(service, conn):
    classifier = None
    label_mapping = None
    auto_archive_settings = None

    cur = conn.cursor()

    try:
        watermark = get_watermark(conn)

        new_watermark = None

        messages_resp = service.users().messages().list(userId='me', q='in:inbox').execute()
        if len(messages_resp['messages']) == 0:
            return

        for message_info in messages_resp['messages']:
            message_id = message_info['id']
            if message_id == watermark.message_id:
                break
            
            message = service.users().messages().get(userId='me', id=message_id).execute()
            if int(message['internalDate']) <= watermark.internal_date:
                break
            
            if new_watermark is None:
                new_watermark = watermark_for_message(message)
            
            contents = get_message_contents(message)
            if classifier is None:
                classifier = get_classifier()

            labels = classifier.classify(contents)
            if len(labels) == 0:
                continue
            
            if label_mapping is None:
                label_mapping = get_labels(service)

            label_ids = []
            for label in labels:
                label_ids.append(label_mapping.id_for_name(label))

            subject = contents.split('\n')[0]
            labels_str = ', '.join(labels)
            print(f'adding labels ({labels_str}) to message {message_id} ({subject})')
            service.users().messages().modify(userId='me', id=message_id, body={'addLabelIds': label_ids}).execute()

            if auto_archive_settings is None:
                auto_archive_settings = get_auto_archive_settings(conn)
            archive_or_queue(service, conn, auto_archive_settings, message_id, labels)

        if new_watermark:
            set_watermark(conn, new_watermark)
    finally:
        cur.close()



import os
import shutil

from .data_dir import get_data_dir
from .gmail import get_creds

def link():
    get_creds()

def unlink():
    dir = get_data_dir()
    if os.path.exists(dir):
        shutil.rmtree(dir)